/* User-defined primitives for C callout from Squeak
 *
 * Author: Boris G. Chr. Shingarov
 *
 * This code was developed in my spare time and has nothing
 * to do with any organization, or company, including, but
 * not limited to, Dialogue-MSU Ltd. and IBM Corp.
 * This code is distributed under the Squeak License.
 *
 * ATTENTION!!!
 * THIS IS EXPERIMENTAL CODE.  IF IT REFORMATS YOUR HARD
 * DRIVE, DAMAGES YOUR HARDWARE, CAUSES DEATH OF YOUR EMPLOYEES,
 * OR WHATEVER OTHER DAMAGE, I'M NOT LIABLE FOR THAT.
 */


#define INCL_DOSMODULEMGR
#define INCL_DOSERRORS
#include <os2.h>

/*
What you basically need to call out to the outside world are two kinds
of things, a DLL, and a CFunction.  A DLL has a name and an hModule.
When you first use your DLL, you must dynamically load the DLL sending
it the #load message.  What will happen is a DosLoadModule() API call
which will return an hModule.

Then, a C Function is a ByteArray of 32 bytes:
*/
typedef struct {
  unsigned address;  /* This will be initially zero.  When you try
                        to call your C function, the callout primitive
                        fails because of this, setting the primitive
                        fail reason to '0000'.  Then the fail code in
                        #call is executed, and that fail code sees
                        if the fail reason was that the address field
                        was 0 (the function was not linked yet).  If
                        so, the fail code will send a #bind (which
                        will do a DosQueryProcAddr()) and resend #call.
                        NB!! In fact, this doesn't work so current
                        release doesn't include fail reasons :-(
                        Instead, this situation is handled in Smalltalk
                        code (querying if the address is valid instead
                        testing if failReason = '0000').
                    */
  unsigned hModule;
  unsigned char flags;
#define SqFnIsNamed 1
	  /* bit  0   SqFnIsNamed
             bits 1-3 retTypeCode
	     bits 4-7 parmCount (0 to 15)
	   */
  unsigned char parmType[15];
  int fn;            /* Functions inside DLLs can be identified by either
                        their ordinal number or their name.  If a function
                        is identified by ordinal number, then fn contains
                        that ordinal number (C value, not oop).  If it's
                        identified by name, fn contains the oop (sic!)
                        of the Smalltalk string with that name.
                     */
  int reserved;      /* I don't want to recreate the DLL dictionaries
                        in the future
                     */
} SqCFunction;

#include "sqUserPrims.h"
#include "CArgTypes.h"




/*************************************************************************
 *             DosLoadModule() / DosFreeModule()                         *
 *************************************************************************/


static int sqDosLoadModule(PHMODULE hModule, int nameIndex, int nameSize);
/* DLL class>>load: aModuleName
 * Return a 32 bit Integer which is the hModule if DosLoadModule() succeeds,
 * fail if the latter fails (set primitive fail reason).
 */
static oop primitiveDosLoadModule(oop receiver, oop moduleName)
{
  unsigned rc, hModule, nameIndex, nameSize;
  if (fetchClassOf(moduleName) != String) return primitiveFail();
  nameIndex = moduleName + 4;
  nameSize = lengthOf(moduleName);
  rc = sqDosLoadModule((PHMODULE)&hModule, nameIndex, nameSize);
  if (rc) return primitiveFail();
  return positive32BitIntegerFor(hModule);
}

/* DLL>>unload: hModule
 * Free hModule.  Return the OS rc, ie., this 
 * primitive does NOT fail if the call fails,
 * but returnsthe reason code.
 */
static oop primitiveDosFreeModule(oop receiver, oop h) {
  return positive32BitIntegerFor( DosFreeModule(positive32BitValueOf(h)) );
}


/* Used internally by primitiveDosLoadModule() */
static int sqDosLoadModule(PHMODULE hModule, int nameIndex, int nameSize)
{
  int rc, i;
  char cDllName[1001];
  char errorString[1000]; /* Not used */

  /* copy the DLL name into a null-terminated C string */
  if (nameSize > 1000) {
	return 123; /* Invalid Name (too long) */
       }
  for (i = 0; i < nameSize; i++) {
	cDllName[i] = *((char *) (nameIndex + i));
       }
  cDllName[nameSize] = 0;

  /* Ok, now we have a null-terminated DLL name in cDllName - load it! */
  rc = DosLoadModule (errorString, 1000, cDllName, hModule);
  return rc;
}



/*************************************************************************
 *                    DosQueryProcAddr () -                              *
 *   called in CFunction>>primDosQueryProcAddr.  The receiver is a C     *
 *   function (essentially, a byte array with structure of SqCFunction). *
 *   The work of this primitive is entirely dedicated to this structure. *
 *   Namely, this primitive fills up the address field of the structure  *
 *   given hModule and fn.  This primitive will not fail due to OS       *
 *   error, but return a non-zero integer instead.  The only thing that  *
 *   may cause this primitive to fail is something that should really    *
 *   never happen, like receiver having the wrong shape.                 *
 *************************************************************************/
static int sqDosQueryProcAddr(oop receiver);
static oop primitiveDosQueryProcAddr(oop receiver) {
  int rc;
  rc = sqDosQueryProcAddr(receiver);
  if (rc==-1) return primitiveFail();
  return positive32BitIntegerFor(rc);
}

static int sqDosQueryProcAddr(oop receiver) {
  SqCFunction *f;
  PFN pfn;
  int rc;

  /*  TO-DO!!!  Add checks if receiver is really a CFunction  */
  f = (SqCFunction*) (receiver+4);
  if (f->hModule==0) return -1;
  if ( (f->flags)&SqFnIsNamed )
     {
       char functionName[256];
       int i, nameSize;

       /* copy the function name into a null-terminated C string */
       nameSize = indexableSize(f->fn);
       if (nameSize > 255) return primitiveFail();
       for (i = 0; i < nameSize; i++) 
                   functionName[i] = *((char *) (f->fn+4+i));
       functionName[nameSize] = 0;
       rc = DosQueryProcAddr (f->hModule, 0, functionName, (PPFN)&pfn);
     }  else
       rc = DosQueryProcAddr (f->hModule, f->fn, NULL, (PPFN)&pfn);
  f->address=(unsigned)pfn;  
  return rc;
}



/*************************************************************************
 *************************************************************************
 *************************************************************************
 *************************************************************************
 *************************************************************************
 *************************************************************************
 ********                  CALLING                                ********
 *************************************************************************
 *************************************************************************
 *************************************************************************
 *************************************************************************
 *************************************************************************
 *************************************************************************/

static oop primitiveCall(oop receiver) {
   SqCFunction *f;
   oop returnOop;
   int (*pf)();
   int retTypeCode;

   f = (SqCFunction*)(receiver+4);
   if ((pf=f->address)==NULL) /* The function was not bound.  This is */
         {                    /* no problem, we'll bind it now, but it */
          int rc;             /* should not normally happen */
          printf("Attempt to call an unbound function - bind it now\n");
          rc = sqDosQueryProcAddr(receiver);
          if (rc!=0) return primitiveFail();
         }
   if (((f->flags>>4)&15)!=0) return primitiveFail(); /* arg count */
   retTypeCode = (f->flags>>1)&7;

   /***  Actual function call  ***/
   returnOop = (*pf) ();

   /***  Convert the return value and pass back to Smalltalk  ***/
   return c2stconversion(returnOop, retTypeCode);
}


static oop primitiveCallW(oop receiver,
	oop arg1) {
   SqCFunction *f;
   oop returnOop;
   int (*pf)(int);
   int parmCount;
   int retTypeCode;
   int x1;

   f = (SqCFunction*)(receiver+4);
   if ((pf=f->address)==NULL) /* The function was not bound.  This is */
         {                    /* no problem, we'll bind it now, but it */
          int rc;             /* should not normally happen */
          printf("Attempt to call an unbound function - bind it now\n");
          rc = sqDosQueryProcAddr(receiver);
          if (rc!=0) return primitiveFail();
         }
   if (((f->flags>>4)&15)!=1) return primitiveFail(); /* arg count */
   retTypeCode = (f->flags>>1)&7;

   /***  Convert the parameters to C  ***/
   x1 = st2cconversion(arg1, f->parmType[0]);

   /***  Actual function call  ***/
   returnOop = (*pf) (x1);

   /***  Careful here.  Some types of conversion (eg. for Strings)
         require memory allocation, ie. st2cconversion() allocates
         a *new* string.  Now, we are going to deallocate these, -
         PLEASE CHECK IF THIS IS WHAT YOU INTENDED  ***/
   if (f->parmType[0]==C_STRING) free((char*)x1);

   /***  Convert the return value and pass back to Smalltalk  ***/
   return c2stconversion(returnOop, retTypeCode);
}





static oop primitiveCallWW(oop receiver,
	oop arg1, oop arg2) {
   SqCFunction *f;
   oop returnOop;
   int (*pf)(int,int);
   int parmCount;
   int retTypeCode;
   int x1,x2;

   f = (SqCFunction*)(receiver+4);
   if ((pf=f->address)==NULL) /* The function was not bound.  This is */
         {                    /* no problem, we'll bind it now, but it */
          int rc;             /* should not normally happen */
          printf("Attempt to call an unbound function - bind it now\n");
          rc = sqDosQueryProcAddr(receiver);
          if (rc!=0) return primitiveFail();
         }
   if (((f->flags>>4)&15)!=2) return primitiveFail(); /* arg count */
   retTypeCode = (f->flags>>1)&7;

   /***  Convert the parameters to C  ***/
   x1 = st2cconversion(arg1, f->parmType[0]);
   x2 = st2cconversion(arg2, f->parmType[1]);

   /***  Actual function call  ***/
   returnOop = (*pf) (x1, x2);

   /***  Careful here.  Some types of conversion (eg. for Strings)
         require memory allocation, ie. st2cconversion() allocates
         a *new* string.  Now, we are going to deallocate these, -
         PLEASE CHECK IF THIS IS WHAT YOU INTENDED  ***/
   if (f->parmType[0]==C_STRING) free((char*)x1);
   if (f->parmType[1]==C_STRING) free((char*)x2);

   /***  Convert the return value and pass back to Smalltalk  ***/
   return c2stconversion(returnOop, retTypeCode);
}


static oop primitiveCallWWW(oop receiver,
	oop arg1, oop arg2, oop arg3) {
   SqCFunction *f;
   oop returnOop;
   int (*pf)(int,int,int);
   int parmCount;
   int retTypeCode;
   int x1,x2,x3;
   f = (SqCFunction*)(receiver+4);
   if ((pf=f->address)==NULL) /* The function was not bound.  This is */
         {                    /* no problem, we'll bind it now, but it */
          int rc;             /* should not normally happen */
          printf("Attempt to call an unbound function - bind it now\n");
          rc = sqDosQueryProcAddr(receiver);
          if (rc!=0) return primitiveFail();
         }
   if (((f->flags>>4)&15)!=3) return primitiveFail(); /* arg count */
   retTypeCode = (f->flags>>1)&7;

   /***  Convert the parameters to C  ***/
   x1 = st2cconversion(arg1, f->parmType[0]);
   x2 = st2cconversion(arg2, f->parmType[1]);
   x3 = st2cconversion(arg3, f->parmType[2]);

   /***  Actual function call  ***/
   returnOop = (*pf) (x1, x2, x3);

   /***  Careful here.  Some types of conversion (eg. for Strings)
         require memory allocation, ie. st2cconversion() allocates
         a *new* string.  Now, we are going to deallocate these, -
         PLEASE CHECK IF THIS IS WHAT YOU INTENDED  ***/
   if (f->parmType[0]==C_STRING) free((char*)x1);
   if (f->parmType[1]==C_STRING) free((char*)x2);
   if (f->parmType[2]==C_STRING) free((char*)x3);

   /***  Convert the return value and pass back to Smalltalk  ***/
   return c2stconversion(returnOop, retTypeCode);
}


static oop primitiveCallWWWW(oop receiver,
	oop arg1, oop arg2, oop arg3, oop arg4) {
   SqCFunction *f;
   oop returnOop;
   int (*pf)(int,int,int,int);
   int parmCount;
   int retTypeCode;
   int x1,x2,x3,x4;
   f = (SqCFunction*)(receiver+4);
   if ((pf=f->address)==NULL) /* The function was not bound.  This is */
         {                    /* no problem, we'll bind it now, but it */
          int rc;             /* should not normally happen */
          printf("Attempt to call an unbound function - bind it now\n");
          rc = sqDosQueryProcAddr(receiver);
          if (rc!=0) return primitiveFail();
         }
   if (((f->flags>>4)&15)!=4) return primitiveFail(); /* arg count */
   retTypeCode = (f->flags>>1)&7;

   /***  Convert the parameters to C  ***/
   x1 = st2cconversion(arg1, f->parmType[0]);
   x2 = st2cconversion(arg2, f->parmType[1]);
   x3 = st2cconversion(arg3, f->parmType[2]);
   x4 = st2cconversion(arg4, f->parmType[3]);

   /***  Actual function call  ***/
   returnOop = (*pf) (x1, x2, x3, x4);

   /***  Careful here.  Some types of conversion (eg. for Strings)
         require memory allocation, ie. st2cconversion() allocates
         a *new* string.  Now, we are going to deallocate these, -
         PLEASE CHECK IF THIS IS WHAT YOU INTENDED  ***/
   if (f->parmType[0]==C_STRING) free((char*)x1);
   if (f->parmType[1]==C_STRING) free((char*)x2);
   if (f->parmType[2]==C_STRING) free((char*)x3);
   if (f->parmType[3]==C_STRING) free((char*)x4);

   /***  Convert the return value and pass back to Smalltalk  ***/
   return c2stconversion(returnOop, retTypeCode);
}


static oop primitiveCallWWWWW(oop receiver,
	oop arg1, oop arg2, oop arg3, oop arg4, oop arg5) {
   SqCFunction *f;
   oop returnOop;
   int (*pf)(int,int,int,int,int);
   int parmCount;
   int retTypeCode;
   int x1,x2,x3,x4,x5;
   f = (SqCFunction*)(receiver+4);
   if ((pf=f->address)==NULL) /* The function was not bound.  This is */
         {                    /* no problem, we'll bind it now, but it */
          int rc;             /* should not normally happen */
          printf("Attempt to call an unbound function - bind it now\n");
          rc = sqDosQueryProcAddr(receiver);
          if (rc!=0) return primitiveFail();
         }
   if (((f->flags>>4)&15)!=5) return primitiveFail(); /* arg count */
   retTypeCode = (f->flags>>1)&7;

   /***  Convert the parameters to C  ***/
   x1 = st2cconversion(arg1, f->parmType[0]);
   x2 = st2cconversion(arg2, f->parmType[1]);
   x3 = st2cconversion(arg3, f->parmType[2]);
   x4 = st2cconversion(arg4, f->parmType[3]);
   x5 = st2cconversion(arg5, f->parmType[4]);

   /***  Actual function call  ***/
   returnOop = (*pf) (x1, x2, x3, x4, x5);

   /***  Careful here.  Some types of conversion (eg. for Strings)
         require memory allocation, ie. st2cconversion() allocates
         a *new* string.  Now, we are going to deallocate these, -
         PLEASE CHECK IF THIS IS WHAT YOU INTENDED  ***/
   if (f->parmType[0]==C_STRING) free((char*)x1);
   if (f->parmType[1]==C_STRING) free((char*)x2);
   if (f->parmType[2]==C_STRING) free((char*)x3);
   if (f->parmType[3]==C_STRING) free((char*)x4);
   if (f->parmType[4]==C_STRING) free((char*)x5);

   /***  Convert the return value and pass back to Smalltalk  ***/
   return c2stconversion(returnOop, retTypeCode);
}


static oop primitiveCallWWWWWW(oop receiver,
	oop arg1, oop arg2, oop arg3, oop arg4, oop arg5, oop arg6) {
   SqCFunction *f;
   oop returnOop;
   int (*pf)(int,int,int,int,int,int);
   int parmCount;
   int retTypeCode;
   int x1,x2,x3,x4,x5,x6;
   f = (SqCFunction*)(receiver+4);
   if ((pf=f->address)==NULL) /* The function was not bound.  This is */
         {                    /* no problem, we'll bind it now, but it */
          int rc;             /* should not normally happen */
          printf("Attempt to call an unbound function - bind it now\n");
          rc = sqDosQueryProcAddr(receiver);
          if (rc!=0) return primitiveFail();
         }
   if (((f->flags>>4)&15)!=6) return primitiveFail(); /* arg count */
   retTypeCode = (f->flags>>1)&7;

   /***  Convert the parameters to C  ***/
   x1 = st2cconversion(arg1, f->parmType[0]);
   x2 = st2cconversion(arg2, f->parmType[1]);
   x3 = st2cconversion(arg3, f->parmType[2]);
   x4 = st2cconversion(arg4, f->parmType[3]);
   x5 = st2cconversion(arg5, f->parmType[4]);
   x6 = st2cconversion(arg6, f->parmType[5]);

   /***  Actual function call  ***/
   returnOop = (*pf) (x1, x2, x3, x4, x5, x6);

   /***  Careful here.  Some types of conversion (eg. for Strings)
         require memory allocation, ie. st2cconversion() allocates
         a *new* string.  Now, we are going to deallocate these, -
         PLEASE CHECK IF THIS IS WHAT YOU INTENDED  ***/
   if (f->parmType[0]==C_STRING) free((char*)x1);
   if (f->parmType[1]==C_STRING) free((char*)x2);
   if (f->parmType[2]==C_STRING) free((char*)x3);
   if (f->parmType[3]==C_STRING) free((char*)x4);
   if (f->parmType[4]==C_STRING) free((char*)x5);
   if (f->parmType[5]==C_STRING) free((char*)x6);

   /***  Convert the return value and pass back to Smalltalk  ***/
   return c2stconversion(returnOop, retTypeCode);
}


static oop primitiveCallWWWWWWW(oop receiver,
	oop arg1, oop arg2, oop arg3, oop arg4, oop arg5, oop arg6, oop arg7) {
   SqCFunction *f;
   oop returnOop;
   int (*pf)(int,int,int,int,int,int,int);
   int parmCount;
   int retTypeCode;
   int x1,x2,x3,x4,x5,x6,x7;
   f = (SqCFunction*)(receiver+4);
   if ((pf=f->address)==NULL) /* The function was not bound.  This is */
         {                    /* no problem, we'll bind it now, but it */
          int rc;             /* should not normally happen */
          printf("Attempt to call an unbound function - bind it now\n");
          rc = sqDosQueryProcAddr(receiver);
          if (rc!=0) return primitiveFail();
         }
   if (((f->flags>>4)&15)!=7) return primitiveFail(); /* arg count */
   retTypeCode = (f->flags>>1)&7;

   /***  Convert the parameters to C  ***/
   x1 = st2cconversion(arg1, f->parmType[0]);
   x2 = st2cconversion(arg2, f->parmType[1]);
   x3 = st2cconversion(arg3, f->parmType[2]);
   x4 = st2cconversion(arg4, f->parmType[3]);
   x5 = st2cconversion(arg5, f->parmType[4]);
   x6 = st2cconversion(arg6, f->parmType[5]);
   x7 = st2cconversion(arg7, f->parmType[6]);

   /***  Actual function call  ***/
   returnOop = (*pf) (x1, x2, x3, x4, x5, x6, x7);

   /***  Careful here.  Some types of conversion (eg. for Strings)
         require memory allocation, ie. st2cconversion() allocates
         a *new* string.  Now, we are going to deallocate these, -
         PLEASE CHECK IF THIS IS WHAT YOU INTENDED  ***/
   if (f->parmType[0]==C_STRING) free((char*)x1);
   if (f->parmType[1]==C_STRING) free((char*)x2);
   if (f->parmType[2]==C_STRING) free((char*)x3);
   if (f->parmType[3]==C_STRING) free((char*)x4);
   if (f->parmType[4]==C_STRING) free((char*)x5);
   if (f->parmType[5]==C_STRING) free((char*)x6);
   if (f->parmType[6]==C_STRING) free((char*)x7);

   /***  Convert the return value and pass back to Smalltalk  ***/
   return c2stconversion(returnOop, retTypeCode);
}








/*************************************************************************
 *************************************************************************
 *************************************************************************
 *          C  <--->  Smalltalk  Conversion  Routines                    *
 *************************************************************************
 *************************************************************************
 *************************************************************************/
/* As these two functions call primitiveFail(), 
   don't use them outside user primitives */
oop c2stconversion(int x, int conversionCode) {
  switch (conversionCode)
  {
   case C_INVALID:
           return primitiveFail();
   case C_VOID:
           return nilObj;
   case C_UINT:
           return positive32BitIntegerFor(x);
   case C_SINT:
   case C_CINT32:
           printf("Still NOT IMPLEMENTED\n");
           return primitiveFail();
   case C_OBJECT:
           return x;
   case C_STRING:
           return (  cString2stString((char*)x)  );
   case C_BYTES:
           printf("Still NOT IMPLEMENTED\n");
           return primitiveFail();
   case C_CHAR:
           printf("Still NOT IMPLEMENTED\n");
           return primitiveFail();
   case C_BOOL:
           return x? trueObj:falseObj;
   default:
           return primitiveFail();
  };
}

int st2cconversion(int x, int conversionCode) {
    char *str;
    int i;
    int len;
  switch (conversionCode)
  {
   case C_INVALID:
           printf("ERROR: C_INVALID passed to a C Function!\n");
           return primitiveFail();
   case C_VOID:
           printf("ERROR: Converting void to a value!\n");
           return primitiveFail();
   case C_UINT:
           return positive32BitValueOf(x);
   case C_SINT:
		printf("Signed integers not implemented!!!\n\n"); exit(66);
   case C_CINT32:
           return longAt(x+4);
   case C_OBJECT:
           return x;
   case C_STRING:
           {
            if(x==1) return 0; /* NULL ptr */
            len = lengthOf(x);
            /* copy the string into a null-terminated C string */
            str = malloc(len + 1);
            for (i = 0; i < len; i++)
         	    str[i] = *((char *) (x + 4/*BHS*/ + i));
             str[len] = 0;
             return ((int)str);
           }
   case C_BYTES: /* either a pointer to ByteArray or NULL */
           if(x==1) return 0; /* NULL ptr */
           if(x&1) return primitiveFail(); /* nonzero integer value - invalid */
           if(isBytes(x)) return x+4;
                 else return primitiveFail(); /* everything else is illegal */
   case C_CHAR:
           return asciiOfCharacter(x);
   case C_BOOL:
           if(x==trueObj) return 1;
           if(x==falseObj) return 0;
           return primitiveFail();
   default:
           printf("ERROR: Unknown C function argument type\n");
           return primitiveFail();
  };
}

oop cString2stString(char *s) {
    char *stStringLocation;
    int i;
    oop o;
    int length;

	length = strlen(s);
	o = instantiateClassindexableSize(String, length);
	stStringLocation = ((char *) (o + 4));
	for (i = 0; i <= (length - 1); i += 1) {
		stStringLocation[i] = (s[i]);
	}
	return o;
}


/*************************************************************************
 *           primitiveGetFailReason - THIS IS BROKEN                     *
 *************************************************************************/
static oop primitiveGetFailReason(oop receiver)
{
 /*
  *   Code for primitiveGetFailReason() is not included
  *   in this release because it's somehow broken with VM 1.23
  */
}




/*************************************************************************
 *                   OBJECT MEMORY ACCESS                                *
 *************************************************************************/

static oop primitiveUint32At(oop receiver, oop i) {
  return positive32BitIntegerFor(
            longAt(receiver+4+positive32BitValueOf(i))
          );
}

static oop primitiveUint32AtPut(oop receiver, oop i, oop x) {
  longAtput(receiver+4+positive32BitValueOf(i), positive32BitValueOf(x));
  return x;
}

static oop primitiveOop32At(oop receiver, oop i) {
  return longAt(receiver+4+positive32BitValueOf(i));
}

static oop primitiveOop32AtPut(oop receiver, oop i, oop x) {
  longAtput(receiver+4+positive32BitValueOf(i), x);
  return x;
}




/*************************************************************************/

char *up_version= "DLL primitives installed";

void sqInitializeUserPrimitives(void)
{
  declarePrimitiveIndexFunctionArgs(1500, primitiveCall, 0);
  declarePrimitiveIndexFunctionArgs(1501, primitiveCallW, 1);
  declarePrimitiveIndexFunctionArgs(1502, primitiveCallWW, 2);
  declarePrimitiveIndexFunctionArgs(1503, primitiveCallWWW, 3);
  declarePrimitiveIndexFunctionArgs(1504, primitiveCallWWWW, 4);
  declarePrimitiveIndexFunctionArgs(1505, primitiveCallWWWWW, 5);
  declarePrimitiveIndexFunctionArgs(1506, primitiveCallWWWWWW, 6);
  declarePrimitiveIndexFunctionArgs(1507, primitiveCallWWWWWWW, 7);

  declarePrimitiveIndexFunctionArgs(1520, primitiveDosLoadModule, 1);
  declarePrimitiveIndexFunctionArgs(1521, primitiveDosFreeModule, 1);
  declarePrimitiveIndexFunctionArgs(1523, primitiveDosQueryProcAddr, 0);

  declarePrimitiveIndexFunctionArgs(1530, primitiveUint32At,    1);
  declarePrimitiveIndexFunctionArgs(1531, primitiveUint32AtPut, 2);
  declarePrimitiveIndexFunctionArgs(1532, primitiveOop32At,    1);
  declarePrimitiveIndexFunctionArgs(1533, primitiveOop32AtPut, 2);
}


